(*
 *  beware IRC services, channeldb.pas
 *  Copyright (C) 2002 Bas Steendijk
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *)

unit channeldb;



interface

uses blinklist,buser,bstuff,bconsts,bcmds,accountcore,chanflagstuff,pgdebugout;

const
  chanlev_auto          =  $1;
  chanlev_banned        =  $2;
  chanlev_master        =  $4;
  chanlev_owner         =  $8;
  chanlev_op            = $10;
  chanlev_voice         = $20;
  chanlev_topic         = $40;
  chanlev_autovoiceonly = $80;



  reggedchanflag_destroying=$1000;
  //reggedchanflag_modeless=  $2000;
  //reggedchanflag_local=     $4000;

  reggedchanflag_nameset=   $8000;

  {schanflagsupported='biklmnopstvr';
  isupportreggedchantypes='+#&';
  sisupportchanflags='b,k,l,imnpstr';
  isupportprefix='(ov)@+';}

type
  {one ban}
  tban=class(tlinklist)
    mask:string;
    umask:string;
    date:integer;
    sender:string;
  end;

  treggedchannel=class(tthing)
    //modeflag:integer;
    flags:integer;
    limit:integer;
    user:tlinklist2;
    banlist:tban;
    bancount:integer;
    topic:string;
    topicts:integer;
    modes:string;
    //topicby:string;
    //topictime:integer;
    //localuser:tplinklist;
    usercount:integer;
    ts:integer;
    flastseents:integer;
    //bottag:array[0..maxserviceunit] of pointer; {4 bytes for each service unit}
    {# of non-deaf users behind this link, used for chat routing}
    welcome : string;
    key     : string;
    autolimit : integer;
    destructor destroy; override;
    function getlastseents : integer;
    property lastseents:integer read getlastseents write flastseents;
  end;

  {reggedchannel-user relationship. prev,next are user's reggedchannel list. prev2,next2 are reggedchannel's user list}
  taccountreggedchan=class(tlinklist2)
    ch:treggedchannel;
    ac:taccount;
    flags:integer; {ops, voice}
    //lu:tplinklist; {local-user struct, if any}

  end;

var
  globalreggedchanlist:tlinklist;
  reggedchannelcount : integer;

var
  chanlevtable_auto          : integer;
  chanlevtable_banned        : integer;
  chanlevtable_master        : integer;
  chanlevtable_owner         : integer;
  chanlevtable_op            : integer;
  chanlevtable_voice         : integer;
  chanlevtable_topic         : integer;
  chanlevtable_autovoiceonly : integer;
const
  maxchanlevtable=7;

  chanlevtable:array[0..maxchanlevtable] of tchanflagtable=(
  (c:'a';flag:chanlev_auto         ;num:@chanlevtable_auto),
  (c:'b';flag:chanlev_banned       ;num:@chanlevtable_banned),
  (c:'m';flag:chanlev_master       ;num:@chanlevtable_master),
  (c:'n';flag:chanlev_owner        ;num:@chanlevtable_owner),
  (c:'o';flag:chanlev_op           ;num:@chanlevtable_op),
  (c:'v';flag:chanlev_voice        ;num:@chanlevtable_voice),
  (c:'t';flag:chanlev_topic        ;num:@chanlevtable_topic),
  (c:'g';flag:chanlev_autovoiceonly;num:@chanlevtable_autovoiceonly)
  );



function isreggedchanprefix(c:char):boolean;
function validreggedchanname(const s:string):boolean;

{return ch for given name, nil if not exists, s does not need to be uppercased}
function findreggedchan(const s:string):treggedchannel;

{
reggedchange the name of a reggedchannel
reggedchannel's names don't reggedchange, but a routine like this
better allows things like a search tree
}
procedure setreggedchanname(ch:treggedchannel;const s:string);

{create user/reggedchannel relation (but don't send anything)  returns the reggedchanuser}
function addusertoreggedchannel(ac:taccount;ch:treggedchannel):taccountreggedchan;

{remove user/reggedchannel relation,
if cu is not nil, it is used instead of a slower seacrh
if reggedchannel became empty, it is destroyed

the user must be on the reggedchannel (no checking done here)
}
procedure deluserfromreggedchannel(ac:taccount;ch:treggedchannel;uc:taccountreggedchan);

{make a new reggedchannel, add it to list}
function createreggedchannel:treggedchannel;

{returns true if us is member of ch. allows parameters to be nil}
function isonreggedchannel(ac:taccount;ch:treggedchannel):boolean;

//function getaccountreggedchan(us:tuser;ch:treggedchannel):taccountreggedchan;
function getaccountreggedchan(ac:taccount;ch:treggedchannel):taccountreggedchan;
{if user can see reggedchannel (in whois etc)}
//function canseereggedchannel(us:tuser;ch:treggedchannel):boolean;
//function issecret(ch:treggedchannel):boolean;
//function isprivate(ch:treggedchannel):boolean;
//function ismoderated(ch:treggedchannel):boolean;
//function isnoexternal(ch:treggedchannel):boolean;

//function hasops(us:tuser;ch:treggedchannel;uc:taccountreggedchan):boolean;
//function hasvoice(us:tuser;ch:treggedchannel;uc:taccountreggedchan):boolean;

//function chanflagstr(ch:treggedchannel;params:boolean):string;

{adds a ban to the reggedchannel. if the ban overlaps existing bans,
theyll be removed and a modereggedchange sent to local users}
procedure addban(sender:tuser;ch:treggedchannel;const s:string);

{returns pointer to ban or nil if the ban is not found}
function findban(ch:treggedchannel;mask:string):tban;

//procedure setchanflag(ch:treggedchannel;cptr,sptr:tuser;parc:integer;parv:pparams);

procedure changechanlev(ac : taccount;rch:treggedchannel;arc:taccountreggedchan;flags : string);
procedure changechanflags(rch:treggedchannel;flags : string);
procedure changechanlevbin(ac : taccount;rch:treggedchannel;arc:taccountreggedchan;flags : integer);
//new is a reserved word in 1.0.x
procedure chanlevstrtobin(ac : taccount;rch:treggedchannel;arc:taccountreggedchan;flags : string;var added,removed,_new : integer);

procedure savechannels;
procedure loadchannels;
procedure clearchannels;

implementation

uses
  sysutils     , bsearchtree , btime , bconfig , channelbot , bparse ,
  unitwritetxt , bcrash      , bchannel;

var
  reggedchantree:thashtable;


function isreggedchanprefix(c:char):boolean;
begin
  result := (c = '#') or (c = '&') or (c = '+');
end;

function validreggedchanname(const s:string):boolean;
var
  a:integer;
begin
  result := false;
  if s = '' then exit;
  if length(s) > maxchanname then exit;
  if not ((s[1] = '#') or (s[1] = '&') or (s[1] = '+')) then exit;
  if length(s) > 1 then for a := 2 to length(s) do begin
    if (s[a] = #7) or (s[a] = ',') or (s[a] = #0) then exit
  end;
  result := true
end;

destructor treggedchannel.destroy;
var
  p:tban;
begin
  if flag_isset(self.flags,reggedchanflag_destroying) then exit;
  setflag(self.flags,reggedchanflag_destroying);

  while self.user <> nil do deluserfromreggedchannel(taccountreggedchan(self.user).ac,self,nil);

  while self.banlist <> nil do begin
    p := self.banlist;
    linklistdel(tlinklist(self.banlist),tlinklist(self.banlist));
    p.destroy;
  end;
  if channelbotobject.up and (ircupper(name)<>ircupper(opt.operchannel)) then sendparse(channelbotobject.mybot.idstr+' '+tok_part+' '+name);
  setreggedchanname(self,'');  {de-init reggedchannel name}
  linklistdel(globalreggedchanlist,tlinklist(self));
  dec(reggedchannelcount);


  inherited destroy;
end;


function findreggedchan(const s:string):treggedchannel;
begin
  result := findtree(@reggedchantree,ircupper(s));
end;

function addusertoreggedchannel(ac:taccount;ch:treggedchannel):taccountreggedchan;
var
  uc:taccountreggedchan;
begin
  uc := taccountreggedchan.create;

  uc.ch := ch;
  uc.ac := ac;

  linklistadd(tlinklist(ac.bottag[channelbotobject.srvunit.num]),tlinklist(uc));
  linklist2add(ch.user,tlinklist2(uc));

  //inc(us.reggedchancount);
  inc(ch.usercount);

  {if myconnect(us) then begin
    lu := tplinklist.create;
    lu.p := us;
    uc.lu := lu;
    linklistadd(tlinklist(ch.localuser),tlinklist(lu));
  end;}
  result := uc;
end;

procedure deluserfromreggedchannel(ac:taccount;ch:treggedchannel;uc:taccountreggedchan);
begin
  if uc = nil then begin
    {search userreggedchan}
    uc := taccountreggedchan(ac.bottag[channelbotobject.srvunit.num]);
    while uc <> nil do begin
      if uc.ch = ch then begin
        break;
      end;
      uc := taccountreggedchan(uc.next);
    end;
  end;

  {if uc.lu <> nil then begin
    linklistdel(tlinklist(ch.localuser),tlinklist(uc.lu));
    uc.lu.destroy
  end;}

  linklistdel(tlinklist(ac.bottag[channelbotobject.srvunit.num]),uc);
  linklist2del(ch.user,uc);
  uc.destroy;
  //dec(us.reggedchancount);
  dec(ch.usercount);

  if ch.usercount <= 0 then begin

    if (ch.flags and reggedchanflag_destroying)=0 then begin
      ch.destroy;
    end;

  end;
end;


function createreggedchannel;
begin
  result := treggedchannel.create;
  result.autolimit := 10;
  linklistadd(globalreggedchanlist,tlinklist(result));
  inc(reggedchannelcount);
end;


procedure setreggedchanname(ch:treggedchannel;const s:string);
begin
  if flag_isset(ch.flags,reggedchanflag_nameset) then deltree(@reggedchantree,ircupper(ch.name));
  ch.name := s;
  if ch.name <> '' then begin
    addtree(@reggedchantree,ircupper(ch.name),ch);
    setflag(ch.flags,reggedchanflag_nameset)
  end else
  clearflag(ch.flags,reggedchanflag_nameset);

end;



function isonreggedchannel(ac:taccount;ch:treggedchannel):boolean;
begin
  result := getaccountreggedchan(ac,ch) <> nil;
end;

function getaccountreggedchan(ac:taccount;ch:treggedchannel):taccountreggedchan;
var
  p2:tlinklist2;
begin
  result := nil;
  if (ch = nil) or (ac = nil) then exit;

  //unfortunately we have only one bottag so we ahve nowhere to store a
  //reggedchancount in the account
  //if ac.reggedchancount <= ch.usercount then begin
    {faster to check user for all reggedchannels}
    p2 := tlinklist2(ac.bottag[channelbotobject.srvunit.num]);
    while p2 <> nil do begin
      if taccountreggedchan(p2).ch = ch then begin
        result := taccountreggedchan(p2);
        exit;
      end;
      p2 := tlinklist2(p2.next);
    end;
  (*
  end else begin
    {faster to check reggedchannel for all users}
    p2 := ch.user;
    while p2 <> nil do begin
      if taccountreggedchan(p2).ch = us then begin
        result := taccountreggedchan(p2);
        exit;
      end;
      p2 := p2.next2;
    end;

  end;*)
end;

{function issecret(ch:treggedchannel):boolean;
begin
  result := flag_isset(ch.modeflag,chanflag_secret);
end;

function isprivate(ch:treggedchannel):boolean;
begin
  result := flag_isset(ch.modeflag,chanflag_private);
end;

function ismoderated(ch:treggedchannel):boolean;
begin
  result := flag_isset(ch.modeflag,chanflag_moderated);
end;

function isnoexternal(ch:treggedchannel):boolean;
begin
  result := flag_isset(ch.modeflag,chanflag_noexternal);
end;

function hasops(us:tuser;ch:treggedchannel;uc:taccountreggedchan):boolean;
begin
  result := false;
  if uc = nil then begin
    uc := taccountreggedchan(us.reggedchannel);
    while uc <> nil do begin
      if uc.ch = ch then begin
        break;
      end;
      uc := taccountreggedchan(uc.next);
    end;
  end;
  if uc = nil then exit;
  result := flag_isset(uc.flags,chanlev_op);
end;
}(*
function hasvoice(us:tuser;ch:treggedchannel;uc:taccountreggedchan):boolean;
begin
  result := false;
  if uc = nil then begin
    uc := taccountreggedchan(us.reggedchannel);
    while uc <> nil do begin
      if uc.ch = ch then begin
        break;
      end;
      uc := taccountreggedchan(uc.next);
    end;
  end;
  if uc = nil then exit;
  result := flag_isset(uc.flags,chanlev_voice);
  {}
end;


function canseereggedchannel(us:tuser;ch:treggedchannel):boolean;
begin
  if not (flag_isset(ch.modeflag,chanflag_secret) or
  flag_isset(ch.modeflag,chanflag_private)) then begin
    result := true;
    exit;
  end;
  result := isonreggedchannel(us,ch);
end;
*)
{
function chanflagstr(ch:treggedchannel;params:boolean):string;
var
  s2:string;
  a:integer;
begin
  if (ch.flags  = 0) and (ch.key = '') and (ch.limit = 0) then begin
    result := '';
    exit;
  end;
  result := ' +';
  s2 := '';
  for a := 0 to maxchanflagtable do begin
    if ch.flags and chanflagtable[a].flag <> 0 then result := result + chanflagtable[a].c;
  end;
  if ch.key <> '' then begin
    result := result + 'k';
    s2 := s2 + ' '+ch.key;
  end;
  if ch.limit > 0 then begin
    result := result + 'l';
    s2 := s2 + ' '+inttostr(ch.limit);
  end;
  if params then result := result + s2;
end;
}
procedure addban(sender:tuser;ch:treggedchannel;const s:string);
var
  s2:string;
  b,b2:tban;
begin
  s2 := ircupper(s);

  {remove overlapped bans}

  b := ch.banlist;
  while b <> nil do begin
    b2 := tban(b.next);
    if maskmatch(s2,b.umask) then begin
      linklistdel(tlinklist(ch.banlist),tlinklist(b));
      b.destroy;
      dec(ch.bancount);
    end;
    b := b2;
  end;

  b := tban.create;
  b.mask := s;
  b.umask := s2;
  b.sender := sender.name;
  b.date := irctime;
  linklistadd(tlinklist(ch.banlist),tlinklist(b));
  inc(ch.bancount);
end;

function findban(ch:treggedchannel;mask:string):tban;
var
  b:tban;
begin
  result := nil;
  mask := ircupper(mask);
  b := ch.banlist;
  while b <> nil do begin
    if b.umask = mask then begin
      result := b;
      exit;
    end;
    b := tban(b.next);
  end;
end;

procedure changechanlev(ac : taccount;rch:treggedchannel;arc:taccountreggedchan;flags : string);
var
  added , removed , _new : integer ;

begin
  if arc=nil then begin
    // we need to find arc
    arc:=getaccountreggedchan(ac,rch);
  end;
  chanlevstrtobin(ac,rch,arc,flags,added,removed,_new);
  changechanlevbin(ac,rch,arc,_new);
end;


procedure chanlevstrtobin(ac : taccount;rch:treggedchannel;arc:taccountreggedchan;flags : string;var added,removed,_new : integer);
var
  removeing : boolean ;
  counter   : byte    ;
  counter2  : byte    ;
  changed   : integer ;
begin

  if arc=nil then begin
    // we need to find arc
    arc:=getaccountreggedchan(ac,rch);
  end;
  if arc=nil then begin
    _new := 0;
  end else begin
    _new := arc.flags;
  end;

  //flags := ircupper(flags);
  removeing := false;
  for counter := 1 to length(flags) do begin
    case flags[counter] of
      '+' : begin
        removeing := false;
      end;
      '-' : begin
        removeing := true;
      end;
      else begin
        for counter2 := 0 to maxchanlevtable do begin
          if chanlevtable[counter2].c = flags[counter] then begin
            if removeing then begin
              _new := _new and (not chanlevtable[counter2].flag);
            end else begin
              _new := _new or chanlevtable[counter2].flag;
            end;
          end;
        end;
      end;
    end;
  end;
  if assigned(arc) then begin
    changed := arc.flags xor _new;
    added := changed and _new;
    removed := changed and arc.flags;
  end else begin
    added := _new;
    removed := 0;
  end;
end;

procedure changechanlevbin(ac : taccount;rch:treggedchannel;arc:taccountreggedchan;flags : integer);
begin
  if arc=nil then begin
    // we need to find arc
    arc:=getaccountreggedchan(ac,rch);
  end;
  if arc=nil then begin
    // we must add user to channel
    arc := addusertoreggedchannel(ac,rch);
  end;
  arc.flags := flags;
  if arc.flags=0 then begin
    deluserfromreggedchannel(ac,rch,arc);
  end;


end;

procedure changechanflags(rch:treggedchannel;flags : string);
var
  removeing : boolean ;
  counter   : byte    ;
  counter2  : byte    ;
begin
  removeing := false;
  for counter := 1 to length(flags) do begin
    case flags[counter] of
      '+' : begin
        removeing := false;
      end;
      '-' : begin
        removeing := true;
      end;
      else begin
        for counter2 := 0 to maxchanflagtable do begin
          if chanflagtable[counter2].c = flags[counter] then begin
            if removeing then begin
              rch.flags := rch.flags and (not chanflagtable[counter2].flag);
            end else begin
              rch.flags := rch.flags or chanflagtable[counter2].flag;
            end;
          end;
        end;
      end;
    end;
  end;
end;



procedure savechannels;
var
  t : text;
  currentreggedchannel : treggedchannel;
  arc                  : taccountreggedchan;
  ch:tchannel;
begin
  openforwrite(t,channelsfile);
  currentreggedchannel := treggedchannel(globalreggedchanlist);
  while assigned(currentreggedchannel) do begin
    ch := findchan(currentreggedchannel.name);
    if assigned(ch) then currentreggedchannel.modes := copy(chanmodestr(ch,true),3,500);
    {dont copy the first 2 chars which are " +" if any modes are set}

    writeln(t,'[channel]');
    writeln(t,'name='+currentreggedchannel.name);
    writeln(t,'flags='+makeflagstr(chanflagtable,maxchanflagtable, currentreggedchannel.flags));
    writeln(t,'limit='+inttostr(currentreggedchannel.limit));
    writeln(t,'topic='+currentreggedchannel.topic);
    writeln(t,'topicts='+inttostr(currentreggedchannel.topicts));
    writeln(t,'modes='+currentreggedchannel.modes);
    writeln(t,'ts='+inttostr(currentreggedchannel.ts));
    writeln(t,'lastseents='+inttostr(currentreggedchannel.lastseents));
    writeln(t,'welcome='+currentreggedchannel.welcome);
    writeln(t,'key='+currentreggedchannel.key);
    writeln(t,'autolimit='+inttostr(currentreggedchannel.autolimit));
    arc := taccountreggedchan(currentreggedchannel.user);
    while assigned(arc) do begin
      writeln(t,'user='+arc.ac.name+':'+makeflagstr(chanlevtable,maxchanlevtable,arc.flags));
      arc := taccountreggedchan(arc.next2);
    end;
    //writeln(t,'limit='+inttostr(currentreggedchannel.limit));
    //writeln(t,'

    // note : ban saveing isn't coded yet as i am waiting for the core to switch
    // to tbanmask (cidr)
    writeln(t);
    currentreggedchannel := treggedchannel(currentreggedchannel.next);
  end;
  closefile(t);
end;

var
  currentreggedchannel : treggedchannel;

procedure loadchannelsproc(sv,sn:string);
var
  separatorpos : integer;
  ac           : taccount;
  accountname  : string;
  a:integer;
  ch:tchannel;
  modes1,modes2:string;
begin
  debugout('processing channels file line sv='+sv+' sn='+sn);
  if sn='' then exit;
  if sv='NAME' then begin
    if (not validchanname(sn)) or (sn[1]<>'#') then begin
      currentreggedchannel := nil;
      exit;
    end;
    currentreggedchannel := createreggedchannel;

    setreggedchanname(currentreggedchannel,sn);
  end else begin
    //assert(currentreggedchannel<>nil,'channelbot','channel details before first channel name');
    if assigned(currentreggedchannel) then begin
      if sv='FLAGS' then begin
        changechanflags(currentreggedchannel,sn);
      end else if sv='LIMIT' then begin;
        currentreggedchannel.limit := strtoint(sn);
      end else if sv='MODES' then begin;
        currentreggedchannel.modes := sn;
      end else if sv='TOPIC' then begin;
        currentreggedchannel.topic := sn;
      end else if sv='TOPIC' then begin;
        currentreggedchannel.topic := sn;
      end else if sv='TOPICTS' then begin;
        currentreggedchannel.topicts := strtointdef(sn,0);
      end else if sv='LASTSEENTS' then begin;
        currentreggedchannel.lastseents := strtointdef(sn,0);
      end else if sv='TS' then begin;
        currentreggedchannel.ts := strtoint(sn);

        if (not assigned(hostservuser)) and (currentreggedchannel.modes <> '') then begin
          a := pos(' ',currentreggedchannel.modes);
          if a = 0 then begin
            modes1 := currentreggedchannel.modes;
            modes2 := '';
          end else begin
            modes1 := copy(currentreggedchannel.modes,1,a-1);
            modes2 := copy(currentreggedchannel.modes,a+1,500);
          end;
        end else begin
          modes1 := '';
          modes2 := '';
        end;
        if modes2 <> '' then modes2 := ' '+modes2;
        bot_join(channelbotobject.mybot,currentreggedchannel.name,'+o'+modes1,channelbotobject.mybot.idstr+modes2,currentreggedchannel.ts);
        if assigned(hostservuser) then begin
          if (currentreggedchannel.flags and chanflag_forcetopic) <> 0 then begin
            if currentreggedchannel.topic <> '' then sendparse(channelbotobject.mybot.idstr+' '+TOK_TOPIC+' '+currentreggedchannel.name+' :'+currentreggedchannel.topic);
          end else begin
            ch := findchan(currentreggedchannel.name);
            if assigned(ch) then currentreggedchannel.topic := ch.topic;
          end;
        end;
      end else if sv='USER' then begin;
        separatorpos := pos(':',sn);
        accountname := copy(sn,1,separatorpos-1);
        ac := taccountUSINGserviceunit(channelbotobject.srvunit).accountcore.findaccount(accountname);
        if ac=nil then begin
          ac := taccountUSINGserviceunit(channelbotobject.srvunit).accountcore.addaccount(accountname,0);
        end;
        changechanlev(ac,currentreggedchannel,nil,copy(sn,separatorpos+1,500));
      end else if sv='WELCOME' then begin
        currentreggedchannel.welcome := sn;
      end else if sv='KEY' then begin
        currentreggedchannel.key := sn;
      end else if sv='AUTOLIMIT' then begin;
        currentreggedchannel.autolimit := strtointdef(sn,10);
      end;
    end;
  end;
end;

procedure loadchannels;
begin
  currentreggedchannel := nil;
  clearchannels;
  readcfg(channelsfile,@loadchannelsproc);
end;

procedure clearchannels;
begin
  while assigned(globalreggedchanlist) do begin
    globalreggedchanlist.free;
  end;
end;

procedure init;
var
  a:integer;
begin
  globalreggedchanlist := nil;
  
  for a := 0 to maxchanlevtable do if chanlevtable[a].num <> nil then
  chanlevtable[a].num^ := a;
end;


function treggedchannel.getlastseents :integer;
var
  arc : taccountreggedchan;
  ac : taccount;
  au : tplinklist;
  us : tuser;
  ch : tchannel;
begin
  result := flastseents;
  ch := findchan(name);
  if not assigned(ch) then exit; // we are nto fully intialised yet
  arc := taccountreggedchan(user);
  while assigned(arc) do begin
    if (arc.flags and chanlev_owner)<>0 then begin
      assert(assigned(ac),'internal structure','no account associated with arc in getlastseents');
      ac := taccount(arc.ac);
      au := ac.user;
      while Assigned(au) do begin
        us := tuser(au.p);
        assert(assigned(ac),'internal structure','no user associated with au in getlastseents');
        if isonchannel(us,ch) then begin
          result := irctime;
          exit;
        end;
        au := tplinklist(au.next);
      end;
    end;
    arc := taccountreggedchan(arc.next2);
  end;
end;

initialization init;


end.

